/*
btgetp.c Version 0.1.0  Btree get record for
the btgetprev routine.
Copyright (C) 2003  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* if record is not FOUND */
/* btgetp returns the next lower record */
/* with a return code of NOTFOUND */

#include <bthash.h>

/* recursive routine */
int btgetp(btreefmt *bthdr, long ofst,
   unsigned char *key, unsigned char *buf)
   {
   int rslt;
   int btgrslt;
   int currcd;
   unsigned char *p,*q,*r;
   unsigned char *rcdptr;
   bt_nodefmt *node;
   bthdr->reason = 0;
   bthdr->rcdnum = 0;
   btgetblk(bthdr,ofst);
   if (bthdr->reason)
      return(BTREE_ERR + 2101);
   rcdptr = (unsigned char *) bthdr->blk + bthdr->blkhdr
      + (bthdr->blk->count * bthdr->rcdnode)
      - bthdr->rcdnode;
   currcd = bthdr->blk->count;
   while (currcd >= 1)
      {
      p = (unsigned char *) rcdptr
	 + bthdr->rcdhdr + bthdr->keyofst;
      q = (unsigned char *) p + bthdr->keylen;
      r = key;
      while (p < q)
	 {
	 rslt = *r - *p;
	 if (rslt) break;
	 p++;
	 r++;
	 } /* for each byte in key */
      if (!rslt)
	 {
	 p = (unsigned char *) rcdptr + bthdr->rcdhdr;
         bthdr->rcd = (unsigned char *) p;
         bthdr->key = (unsigned char *) p + bthdr->keyofst;
         bthdr->dletbyte = (unsigned char *) p + bthdr->dletofst;
	 bthdr->rcdnum = currcd;
	 if (*(p + bthdr->dletofst) == DELETED)
	    {
            bthdr->reason = BTREE_ERR + 2105;
            return(NOTFOUND);
	    } /* deleted record matches key */
         if (buf != NULL)
            {
            q = (unsigned char *) p + bthdr->rcdlen;
            r = (unsigned char *) buf;
            while (p < q) *r++ = *p++;
            } /* if buf specified */
         node = (bt_nodefmt *) rcdptr;
	 if (node->left)
	    bthdr->reason = INDXNODE;
	 else
	    bthdr->reason = LEAFNODE;
	 return(FOUND);
	 } /* if match */
      else if (rslt < 0)
	 {
	 if (currcd > 1)
	    {
	    rcdptr = (unsigned char *) rcdptr - bthdr->rcdnode;
	    currcd--;
	    continue;
	    } /* if not beginning of block */
	 else /* if beginning of block */
	    {
            node = (bt_nodefmt *) rcdptr;
	    if (node->left)
	       {
               return(btgetp(bthdr,node->left,key,buf));
	       } /* if left pointer */
	    else /* beginning of block && no left pointer */
	       {
               if (buf != NULL)
                  {
                  p = (unsigned char *) buf;
                  q = (unsigned char *) buf + bthdr->rcdlen;
                  while (p < q) *p++ = '\0';
                  } /* if buf specified */
	       bthdr->rcdnum = 0;
	       bthdr->reason = FRSTPAPA;
	       return(FRSTPAPA);
	       } /* if not found */
	    } /* if beginning of block */
	 } /* if less than */
      else if (rslt > 0)
	 {
	 if (currcd >= bthdr->blk->count
	    && bthdr->blk->right)
	    {
            btgrslt = btgetp(bthdr,bthdr->blk->right,key,buf);
	    if (btgrslt == FRSTPAPA)
	       {
               btgetblk(bthdr,ofst);
               if (bthdr->reason)
                  return(BTREE_ERR + 2102);
               r = (unsigned char *) rcdptr + bthdr->rcdhdr;
               bthdr->rcd = (unsigned char *) r;
               if (buf != NULL)
                  {
                  p = (unsigned char *) buf;
                  q = (unsigned char *) buf + bthdr->rcdlen;
                  while (p < q) *p++ = *r++;
                  } /* if buf specified */
               bthdr->reason = FRSTPAPA;
               bthdr->rcdnum = currcd;
               return(NOTFOUND);
               } /* if get first parent to right */
	    else return(btgrslt);
	    } /* if end of block and right pointer */
	 else if (currcd >= bthdr->blk->count)
	    {
            r = (unsigned char *) rcdptr + bthdr->rcdhdr;
            bthdr->rcd = (unsigned char *) r;
            if (buf != NULL)
               {
               p = (unsigned char *) buf;
               q = (unsigned char *) buf + bthdr->rcdlen;
               while (p < q) *p++ = *r++;
               } /* if buf specified */
            bthdr->reason = BTEOF;
            bthdr->rcdnum = currcd;
            return(NOTFOUND);
	    } /* if end of block and no right pointer */
	 /* else not end of block and left pointer */
	 /* in next higher record */
         p = (unsigned char *)
	    rcdptr + bthdr->rcdnode;
         node = (bt_nodefmt *) p;
	 if (node->left)
	    {
            btgrslt = btgetp(bthdr,node->left,key,buf);
	    if (btgrslt == FRSTPAPA)
	       {
               btgetblk(bthdr,ofst);
               if (bthdr->reason)
                  return(BTREE_ERR + 2103);
               r = (unsigned char *) rcdptr + bthdr->rcdhdr;
               bthdr->rcd = (unsigned char *) r;
               if (buf != NULL)
                  {
                  p = (unsigned char *) buf;
                  q = (unsigned char *) buf + bthdr->rcdlen;
                  while (p < q) *p++ = *r++;
                  } /* if buf specified */
               bthdr->reason = FRSTPAPA;
               bthdr->rcdnum = currcd;
               return(NOTFOUND);
               } /* if get first parent to left */
	    else return(btgrslt);
	    } /* if left pointer */
	 else
	    {
	    p = (unsigned char *) rcdptr + bthdr->rcdhdr;
            bthdr->rcd = (unsigned char *) p;
            if (buf != NULL)
               {
               q = (unsigned char *) p + bthdr->rcdlen;
               r = (unsigned char *) buf;
               while (p < q) *r++ = *p++;
               } /* if buf specified */
	    bthdr->rcdnum = currcd;
	    bthdr->reason = LEAFNODE;
	    return(NOTFOUND);
	    } /* if not found */
	 } /* if less than */
      } /* for each record in node */
   bthdr->reason = FRSTPAPA;
   return(FRSTPAPA);
   } /* btgetp */
